from __future__ import print_function

from datetime import datetime
from warnings import warn
from socket import AF_INET, SOCK_STREAM, socket
import argparse
import os
from pathlib import Path
import traceback
import shlex
import subprocess
import sys

from . import settings
from .diskemptysampler import DiskEmptySampler
from .hdddock import BaseStorageDevice
from .inventoryrestclient import InventoryRESTClient
from .labelprinterapiclient import get_labelprinter


S_UNSET = '(unset)'


class Inventory:
    def __init__(self, inventory_rest_client, hdd_id):
        self._client = inventory_rest_client
        self._hdd_id = hdd_id
        self.flush()

    def flush(self):
        self._data = self._client.get_hdd(self._hdd_id)[0]

    @property
    def id(self):
        return self._data['id']

    @property
    def tag_uid(self):
        return self._data['tag_uid']

    @property
    def bay(self):
        return self._data['bay']

    @property
    def current_owner(self):
        return (self._data.get('current_owner') or {}).get('name')

    @property
    def current_health_status(self):
        return (self._data.get('current_health') or {}).get('status')


class SymlinkNotFound(Exception):
    pass


class DeviceManager(object):
    def __init__(self, hdd_dock, hdd_id, inventory_rest_client,
                 author, location):
        self._hdd_dock = hdd_dock
        self._hdd_id = hdd_id
        self._inventory_rest_client = inventory_rest_client
        self._author = author
        self._location = location

    def show_summary(self, print_data):
        do_print(print_data, self._hdd_dock)

    def show_commands(self):
        erase_methods = self._hdd_dock.get_erase_methods()
        next_erase_method = 0

        print('')
        print('Inventory-url: {}'.format(
            self._inventory_rest_client.get_hdd_url(self._hdd_id)))
        print('')
        print('Possible actions:')
        print('1. Change owner')
        print('2. Reprint label')
        print('3. Quick erase')
        if len(erase_methods) > 1:
            print('4. Secure erase ({})'.format(
                erase_methods[next_erase_method].name))
            next_erase_method += 1
        print('5. Change server bay')
        print('6. Change health status')
        print('7. Print health label')
        print('8. Secure erase ({})'.format(
            erase_methods[next_erase_method].name))
        print('D. Dispose HDD in security container')
        print('P. Show current info/status (again)')

        print('')
        print('9. Quit + EJECT (^C to skip eject)')

    def _set_status(self, status, extra_info):
        self._inventory_rest_client.add_status(
            self._hdd_id, status=status, extra_info=extra_info)

    def _set_health(self, title):
        # title = 'SECURE ERASED' or ...
        health = (
            'OK' if self._hdd_dock.hwdata.wear_health_percent >= 100
            else '{}%'.format(self._hdd_dock.hwdata.wear_health_percent))
        health_status = f'health {health} - {title}'
        self.set_health(health_status)

    def set_health(self, health_status):
        self._inventory_rest_client.add_health_status(
            self._hdd_id, status=health_status,
            extra_info='--{}'.format(self._author))

    def set_health_with_wear(self, title):
        self._set_health(title)

    def quick_erase(self):
        print('Start quick wiping')
        self._hdd_dock.quick_erase()
        self._set_status('QUICK_WIPED', 'Quick wiped at {} --{}'.format(
            self._location, self._author))
        try:
            self._hdd_dock.flush()
        finally:
            self._set_health('QUICK WIPED')
        print('Disk/ssd quick wiped')

    def best_erase(self):
        erase_method = self._hdd_dock.get_erase_methods()[0]
        self._erase(erase_method)

    def second_best_erase(self):
        erase_methods = self._hdd_dock.get_erase_methods()
        if len(erase_methods) > 1:
            erase_methods.pop(0)
        self._erase(erase_methods[0])

    def _erase(self, erase_method):
        if erase_method.type == 'builtin':
            return self._erase_with_status(erase_method)
        if erase_method.type == 'manual':
            print('Secure disk wipe chosen. Please wait as this will '
                  'take a long time... - 1 pass lessrandom and 1 pass '
                  'zerofill\n')
            return self._erase_with_status(erase_method)
        raise NotImplementedError(erase_method)

    def _erase_with_status(self, erase_method):
        # Fetch vars.
        # XXX: better dynamic value for sample count based on disk size?
        sample_count = settings.POST_WIPE_SAMPLE_COUNT
        if self._hdd_dock.is_ssd():
            sample_count *= 10  # SSDs/NVMes are fast(er)
        sample_size = settings.POST_WIPE_SAMPLE_SIZE

        print('Attempting erase method {!r} on BLKDEV {!r}'.format(
            erase_method.name, self._hdd_dock.devname))

        # Running pre-sample.  This is not needed, but nice during dev.
        print(
            'DEBUG: Sampling block device to check if empty... '
            '(sample_count: {}, sample_size {})'.format(
                sample_count, sample_size))
        sampler = DiskEmptySampler(
            self._hdd_dock.devname, sample_count, sample_size)
        sampler.sample()
        if sampler.is_zero():
            print('DEBUG: No non-zero sample found. Disk already zeroed?')
        else:
            print('DEBUG: Found non-zero samples before wipe (expected)')

        print('Start secure wiping BLKDEV')
        success, error = erase_method()
        if not success:
            print('Error: {}'.format(error))
            self._set_status(
                'BLKDEV_SECURE_WIPED_ERROR', (
                    'Block device secure wipe error at {}: {} --{}'
                    .format(self._location, error, self._author)))
            self._set_health('ERASE ERROR')
            return

        print('Block device securely wiped')

        # Running post-sample. Should be all zero.
        print('DEBUG: Sampling block device afterwards...')
        sampler.sample()
        if sampler.is_zero():
            print('OK: All samples are empty')
        elif sampler.is_different():
            print(
                '(probably) OK: All samples are different '
                '(old school crypto disk?)')
        else:
            print('ERROR: Found not-wiped sample!')
            print('Please check {!r} manually!'.format(self._hdd_dock.devname))
            self._set_status(
                'BLKDEV_SECURE_WIPED_ERROR', (
                    'Block device secure wipe {!r} error at {}, non-empty '
                    'samples found --{}'
                    .format(erase_method.name, self._location, self._author)))
            self._set_health('ERASE ERROR')
            return

        # All good!
        self._set_status(
            'BLKDEV_SECURE_WIPED', (
                'Block device secure wiped using {!r} at {} and '
                'checked {}x{} --{}'
                .format(
                    erase_method.name, self._location, sample_count,
                    sample_size, self._author)))
        try:
            self._hdd_dock.flush()
        finally:
            self._set_health('SECURE ERASED')

    def _find_non_zero_blocks(self, sample_count, sample_size):

        disk_empty_sampler = DiskEmptySampler(
            self._hdd_dock.devname, sample_count, sample_size)
        result = disk_empty_sampler.sample_disk()
        assert result is not None

        return result is False  # false = we found non-zero


def register_disk(**kwargs):
    # Python2-compatible forced kwargs. Temporary fix until we clean up this
    # code.
    hdd_dock = kwargs.pop('hdd_dock')
    inventory_rest_client = kwargs.pop('inventory_rest_client')
    author = kwargs.pop('author')
    location = kwargs.pop('location')
    assert not kwargs, None

    print('')
    print('Disk is not registered yet, please specify the following fields:')
    owner = input('Owner [OSSO]: ')

    # Default to OSSO
    if owner in (None, ''):
        owner = 'OSSO'

    bay = input('bay ['']: ')

    if bay in (None, ''):
        bay = ''

    erase = None
    while erase not in ('y', 'n', 'Y', 'N', ''):
        erase = input('Quick erase? (y/N): ')

    result = inventory_rest_client.add_hdd(hdd_dock, bay)
    tag_uid = result['tag_uid']
    hdd_id = result['id']

    inventory_rest_client.add_smart_data(
        hdd_id, rawdata=hdd_dock.hwdata.rawdata)
    inventory_rest_client.add_status(
        hdd_id, status='REGISTERED',
        extra_info='Registered at {} --{}'.format(location, author))
    inventory_rest_client.add_owner(  # XXX: author-of-owner, not owner-eml
        hdd_id, name=owner, email=author)
    inventory_rest_client.add_location(
        hdd_id, location=location)

    if erase == 'y' or erase == 'Y':
        manager = DeviceManager(
            hdd_dock, hdd_id, inventory_rest_client, author, location)
        manager.quick_erase()

    # Print label, unless the disk is in a remote machine
    if location != 'remote':
        labelprinter_rest_client = get_labelprinter()
        labelprinter_rest_client.print_hdd_label(
            tag_uid, hdd_dock.get_serial_number(), owner)

    input('Registration complete, press enter')


def registered_disk_actions(action=None, **kwargs):
    # Python2-compatible forced kwargs. Temporary fix until we clean up this
    # code.
    hdd_dock = kwargs.pop('hdd_dock')
    hdd_id = kwargs.pop('hdd_id')
    inventory_rest_client = kwargs.pop('inventory_rest_client')
    static_data = kwargs.pop('static_data')
    del static_data
    dynamic_data = kwargs.pop('dynamic_data')
    author = kwargs.pop('author')
    location = kwargs.pop('location')
    assert not kwargs, None

    if not settings.DEBUG:
        # Always add smart data, even when disk is registered
        inventory_rest_client.add_smart_data(
            hdd_id, rawdata=hdd_dock.hwdata.rawdata)
        # Always add a status & location
        # of the disk being seen at OSSO HQ
        if location != 'remote':
            inventory_rest_client.add_status(
                hdd_id, status='CHECKED_IN',
                extra_info='Checked in at {} --{}'.format(location, author))
            inventory_rest_client.add_location(
                hdd_id, location=location)

    inventory = Inventory(inventory_rest_client, hdd_id)

    def _build_print_data():
        static_data, dynamic_data = build_hdd_info(hdd_dock)
        print_data = static_data + dynamic_data
        print_data.append(['', None])
        print_data.append(['REGISTRATION INFORMATION', None])
        print_data.append(['=', None])
        print_data.append(['Disk is already registered as', inventory.id])
        print_data.append(['Last owner', inventory.current_owner or S_UNSET])
        print_data.append([
            'Last health status', inventory.current_health_status or S_UNSET])
        print_data.append(['Server bay', inventory.bay or S_UNSET])
        return print_data

    manager = DeviceManager(
        hdd_dock, hdd_id, inventory_rest_client, author, location)
    manager.show_summary(_build_print_data())

    if action == 'secure-erase':
        print('Automatic: Starting secure erase')
        manager.best_erase()
        inventory.flush()

    manager.show_commands()

    while True:
        action = input('\nAction: ').upper()

        if action == '1':
            owner = ''
            while len(owner) == 0:
                owner = input('New owner: ')
            inventory_rest_client.add_owner(  # XXX: not email-of-owner
                hdd_id, name=owner, email=author)
            inventory.flush()

        elif action == '2':
            # Print label
            labelprinter_rest_client = get_labelprinter()
            labelprinter_rest_client.print_hdd_label(
                inventory.tag_uid, hdd_dock.get_serial_number(),
                inventory.current_owner)

        elif action == '3':
            manager.quick_erase()
            inventory.flush()

        elif action == '4':
            manager.best_erase()
            inventory.flush()

        elif action == '5':
            bay = ''
            while len(bay) == 0:
                bay = input('New server bay: ')
            inventory_rest_client.change_bay(hdd_id, bay)
            inventory.flush()
            print('Server bay changed to: {}'.format(inventory.bay))

        elif action == '6':
            health_status = (
                input('Manual health status [ONLINE]: ') or 'ONLINE')
            manager.set_health_with_wear(health_status)
            inventory.flush()

        elif action == '7':
            # Print label
            print_health_label(hdd_dock, inventory, author, dynamic_data)

        elif action == '8':
            manager.second_best_erase()
            inventory.flush()

        elif action == '9':
            break  # Quit + EJECT

        elif action == 'D':
            assert location != 'remote', 'Cannot remote-dispose..'
            inventory_rest_client.add_status(
                hdd_id, status='HDD_DISPOSED',
                extra_info='Disposed in security container at {} --{}'.format(
                    location, author))
            inventory_rest_client.add_location(
                hdd_id, location='OSSO HQ: HDD security container')
            inventory.flush()
            break

        elif action == 'P':
            manager.show_summary(_build_print_data())
            manager.show_commands()

        else:
            print('Notice: {!r} action unknown'.format(action))

    hdd_dock.shutdown_disk()
    sys.exit(0)


def human_readable_bytes(byte_count):
    options = ['Bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB']

    counter = 0
    while byte_count // 1024 > 0:
        byte_count = byte_count / 1024.0
        counter += 1
        if counter == len(options) - 1:
            break

    return '{} {}'.format(round(byte_count, 2), options[counter])


def do_print(data, hdd_dock):
    key_length = 0
    item_length = 0

    for key, item in data:
        key_length = max(key_length, len(key))
        item_length = max(item_length, len(item or ''))
    heading = '=' * (key_length + item_length + 3)
    fmt = '{:%d} : {}' % (key_length,)

    print('DISK INFORMATION [BAY NR: {}]'.format(
        hdd_dock.docktool_bay_nr))
    print(heading)

    for key, item in data:
        if item is None and key == '=':
            print(heading)
        elif item is None:
            print(key)
        else:
            print(fmt.format(key, item))


def print_health_label(hdd_dock, inventory, author, dynamic_data):
    """
    Print health status on a label. See also print_health_label_from_db.
    Both would like some refactoring.
    """
    today = datetime.now().strftime('%Y-%m-%d')
    # XXX: don't use dynamic_data please :(
    total_bytes_written = [
        x for x in dynamic_data
        if 'Total bytes written' in x[0]][0]
    total_bytes_read = [
        x for x in dynamic_data
        if 'Total bytes read' in x[0]][0]

    lines = [
        'Health status : {}'.format(
            inventory.current_health_status.upper()),
        '',
        'Serial : {}'.format(hdd_dock.get_serial_number()),
        'Total bytes written/read : {}/{}'.format(
            total_bytes_written[1], total_bytes_read[1]),
        'Power on hours : {} hours'.format(
            hdd_dock.hwdata.power_on_hours or S_UNSET),
        f'--{author} @ {today}',
    ]
    try:
        labelprinter_rest_client = get_labelprinter()
        labelprinter_rest_client.print_generic_label(lines)
    except Exception:
        print('Would have printed:')
        print('\n'.join('- {}'.format(line) for line in lines))
        print()
        raise


def print_health_label_from_db(hdd_data):
    """
    A shabby version of print_health_label() using all info we have
    available in the DB.
    """
    print(
        'WARNING: total bytes read/written and total power hours '
        'not known when printing; you should have printed a health label '
        'in "dev" mode', file=sys.stderr)
    print('... but printing a partial label for you anyway.', file=sys.stderr)

    health = hdd_data['current_health']['status'].upper()
    author_of_health = hdd_data['current_health']['extra_info'].strip()
    if '--' in author_of_health:
        author_of_health = author_of_health.rsplit('--', 1)[-1]
    else:
        author_of_health = 'unknown'
    serial = hdd_data['serial_number']
    date_of_health = hdd_data['current_health']['timestamp'].split('T')[0]

    lines = [
        f'Health status : {health} @ {date_of_health}',
        '',
        f'Serial : {serial}',
        'Total bytes written/read : (not known in db)',
        'Power on hours : (not known in db)',
        f'--{author_of_health} @ {date_of_health}',
    ]
    try:
        labelprinter_rest_client = get_labelprinter()
        labelprinter_rest_client.print_generic_label(lines)
    except Exception:
        print('Would have printed:')
        print('\n'.join('- {}'.format(line) for line in lines))
        print()
        raise


def build_hdd_info(hdd_dock):
    hwdata = hdd_dock.hwdata

    static_data = []

    static_data.append(['Device model', hdd_dock.get_device_model()])
    static_data.append(['Serial', hdd_dock.get_serial_number()])
    static_data.append(
        ['Device (port)', '{} ({})'.format(
            hdd_dock.devname, hdd_dock.port if hdd_dock.port else 'Unknown')])
    static_data.append(['SSD', ('yes' if hdd_dock.is_ssd() else 'no')])
    static_data.append(
        ['SAS (detected)', ('yes' if hdd_dock.is_sas() else 'no')])
    static_data.append(['User Capacity', hdd_dock.get_user_capacity()])

    dynamic_data = []

    total_bytes_written = S_UNSET
    total_bytes_read = S_UNSET
    if hwdata.sector_size is not None:
        if hwdata.lbas_written is not None:
            total_bytes_written = human_readable_bytes(
                hwdata.sector_size * hwdata.lbas_written)
        if hwdata.lbas_read is not None:
            total_bytes_read = human_readable_bytes(
                hwdata.sector_size * hwdata.lbas_read)
    dynamic_data.append(['Total bytes written', total_bytes_written])
    dynamic_data.append(['Total bytes read', total_bytes_read])

    def _NS(value):
        if value is None:
            return S_UNSET
        return str(value)

    dynamic_data.append([
        'Power on hours',
        '{} hours'.format(_NS(hwdata.power_on_hours))])
    dynamic_data.append([
        'Wear health percent', '{}% (smart: {})'.format(
            _NS(hwdata.wear_health_percent), hwdata.smart_status)])
    dynamic_data.append([
        'Reallocated sector count',
        _NS(hwdata.reallocated_sector_ct)])
    dynamic_data.append([
        'Reallocated event count',
        _NS(hwdata.reallocated_event_count)])
    dynamic_data.append([
        'Current pending sector',
        _NS(hwdata.current_pending_sector)])
    dynamic_data.append([
        'Offline uncorrectable',
        _NS(hwdata.offline_uncorrectable)])

    return static_data, dynamic_data


def get_author_and_location():
    author = os.environ.get('EMAIL', '')
    if '@' not in author:
        print('ERROR: Please set the EMAIL envvar to specify who you are!')
        sys.exit(1)
    location = os.environ.get('LOCATION', 'remote')
    if not location:
        print('ERROR: Please (un)set the LOCATION envvar to specify where!')
        sys.exit(1)
    return author, location


def dev_menu(devname, action):
    if os.getuid() != 0:
        warn('Expected UID 0 (root) for access')
    author, location = get_author_and_location()
    hdd_dock = BaseStorageDevice.from_devname(devname)
    sys.stdout.write('\x1b]2;DOCKTOOL DISK BAY: {}\x07'.format(
        hdd_dock.docktool_bay_nr))

    # XXX: build_hdd_info() is a quick hack..
    static_data, dynamic_data = build_hdd_info(hdd_dock)

    inventory_rest_client = InventoryRESTClient(
        settings.DASHBOARD_BASE_URL)

    # HDD is an asset so asset_id = hdd_id
    try:
        hdd_id = inventory_rest_client.get_hdd_id(
            hdd_dock.get_device_model(),
            hdd_dock.get_serial_number())
    except ValueError as e:
        warn(str(e))
        do_print(static_data + dynamic_data, hdd_dock)
        raise

    print(
        '\n\x1b[1mBEWARE:\x1b[0m All your actions will be recorded '
        'as performed by: \x1b[1;32m{}\x1b[0m\n'
        'Change EMAIL envvar if it is incorrect!'.format(author))
    print(
        '\x1b[1mBEWARE:\x1b[0m LOCATION is set to: \x1b[1;32m{}\x1b[0m\n'
        'Change to "OSSO HQ: HDD docktool" if in the office.\n'
        .format(location))

    if hdd_id is None:
        do_print(static_data + dynamic_data, hdd_dock)
        register_disk(
            hdd_dock=hdd_dock,
            inventory_rest_client=inventory_rest_client,
            author=author, location=location)
        hdd_id = inventory_rest_client.get_hdd_id(
            hdd_dock.get_device_model(),
            hdd_dock.get_serial_number())
        assert hdd_id, 'HDD id not set after registration?'

    registered_disk_actions(
        hdd_dock=hdd_dock, hdd_id=hdd_id,
        inventory_rest_client=inventory_rest_client,
        static_data=static_data, dynamic_data=dynamic_data,
        author=author, location=location, action=action)


def db_menu(serial):
    inventory_rest_client = InventoryRESTClient(
        settings.DASHBOARD_BASE_URL)

    # HDD is an asset so asset_id = hdd_id
    hdd_id = inventory_rest_client.get_hdd_id(None, serial)  # serial?
    if hdd_id is not None:
        hdd = inventory_rest_client.get_hdd(hdd_id)[0]
    else:
        hdd = inventory_rest_client.get_hdd(serial)[0]  # hdd_id as "serial"
    if hdd is None:
        raise ValueError('serial/asset not found in remote DB')

    # Ok, we have an ID. Get info from remote?
    do_print_health_label = None
    while do_print_health_label not in ('y', 'n', ''):
        do_print_health_label = input('Print health label? (y/N): ').lower()

    if do_print_health_label == 'y':
        print_health_label_from_db(hdd)
        print('done.')
    else:
        from pprint import pprint
        pprint(hdd)


class AutoServer:
    def __init__(self, action, gui, ignore_devices):
        self.action = action
        self.gui = gui
        self.ignore_devices = [i.resolve() for i in ignore_devices]
        # Make sure the environment is valid to start osso-docktool.
        self.author, self.location = get_author_and_location()
        if not gui:
            if not os.environ.get('TMUX'):
                os.execvp('tmux', [
                    'tmux', 'new', '-s', 'docktool', '-n', 'main',
                    f'{self.docktool_exec} auto {action}'])
            else:
                # Add the environment to the tmux session.
                subprocess.check_call(
                    ['tmux', 'set', 'remain-on-exit', 'on'])
                subprocess.check_call(
                    ['tmux', 'setenv', 'EMAIL', self.author])
                subprocess.check_call(
                    ['tmux', 'setenv', 'LOCATION', self.location])

    def start(self):
        try:
            self.listen()
        except KeyboardInterrupt:
            raise SystemExit()

    def listen(self):
        with socket(AF_INET, SOCK_STREAM) as sock:
            sock.bind(('127.0.0.1', 1451))
            sock.listen(4)
            sock.setblocking(True)
            print('osso-docktool automatic™')
            print(
                '\x1b[1mBEWARE:\x1b[0m All your actions will be recorded '
                f'as performed by: \x1b[1;32m{self.author}\x1b[0m\n'
                'Change EMAIL envvar if it is incorrect!')
            print(
                '\x1b[1mBEWARE:\x1b[0m LOCATION is set to: \x1b[1;32m'
                f'{self.location}\x1b[0m\n'
                'Change to "OSSO HQ: HDD docktool" if in the office.')
            print('Action:', self.action)
            if self.action == 'secure-erase':
                print('\x1b[1mBEWARE:\x1b[0m this will destroy the data on '
                      'newly connected disks without confirmation !')
            if self.ignore_devices:
                print(
                    'Ignoring devices:',
                    ','.join([str(i) for i in self.ignore_devices]))
            print('Waiting for disk notifications...')
            while True:
                connection, address = sock.accept()
                pid = os.fork()
                if pid == 0:
                    self.handle_client(connection)
                else:
                    os.waitid(os.P_PID, pid, os.WEXITED)

    def handle_client(self, connection):
        client_data = b''
        with connection:
            while True:
                data = connection.recv(1024)
                if not data:
                    break
                client_data += data

        if client_data:
            device = Path(client_data.decode())
            if device.is_block_device():
                if self.is_ignored(device):
                    print(f'Ignoring notification for {device}')
                else:
                    self.start_docktool(device)
            else:
                print(f'Path {device} is not a block device')

    @property
    def docktool_exec(self):
        return Path(sys.argv[0]).resolve()

    def is_ignored(self, device):
        # Check the resolved path against the ignore list.
        return bool(device.resolve() in self.ignore_devices)

    def start_docktool(self, device):
        if self.gui:
            self.start_gui(device)
        else:
            self.start_cli(device)

    def start_gui(self, device):
        print(f'Starting GUI terminal for {device}')
        # Double fork to detach terminal from pid group.
        if os.fork() == 0:
            args = [
                f'osso-docktool-terminal [{device}]', '-e',
                'sudo', '-E', self.docktool_exec, 'dev', device, '--wait']
            if self.action != 'start':
                args += ['--action', self.action]
            os.setsid()
            os.execvp('x-terminal-emulator', args)
        else:
            os._exit(0)

    def start_cli(self, device):
        print(f'Creating new tmux window for {device}')
        if not os.environ.get('TMUX'):
            print('Error: TMUX environment variable is empty', file=sys.stderr)
        else:
            args = ['tmux', 'new-window', '-n', device]
            cmd = (
                f'sudo -E {self.docktool_exec} dev {shlex.quote(str(device))} '
                '--wait')
            if self.action != 'start':
                cmd += f' --action {self.action}'
            subprocess.check_call(args + [cmd])
        os._exit(0)


def auto_notify(device):
    with socket(AF_INET, SOCK_STREAM) as sock:
        try:
            sock.connect(('127.0.0.1', 1451))
            sock.sendall(str(device).encode())
        except ConnectionError as e:
            print(f'Error notifying auto process: {e}', file=sys.stderr)


class DocktoolArgumentParser(argparse.ArgumentParser):
    def __init__(self):
        super().__init__(
            prog='osso-docktool',
            description='Docktool for processing disks')  # exit_on_error=False
        self.message2 = None

        self.subparsers = self.add_subparsers(
            title='subcommands', dest='command', required=True,
            help='subcommand help',
            parser_class=argparse.ArgumentParser)

        devparser = self.subparsers.add_parser(
            'dev',
            help='(implied!) process hardware device (original behaviour)')
        devparser.add_argument(
            'device', metavar='DISK', type=self.block_device,
            help='Device to use for example: /dev/sda')
        devparser.add_argument(
            '-a', '--action', choices=['secure-erase'],
            help='Automatically execute the action.')
        devparser.add_argument(
            '-w', '--wait', action='store_true',
            help='Wait for user confirmation before exit.')

        dbparser = self.subparsers.add_parser(
            'db', help='process devices device info without hardware access')
        dbparser.add_argument(
            'serial', metavar='SERIAL',
            help='Serial number (or GoCollect asset UID')

        autoparser = self.subparsers.add_parser(
            'auto',
            help='Automatically perform action on newly inserted disks')
        autoparser.add_argument(
            '-g', '--gui', action='store_true',
            help='Start docktool processes in a GUI instead of tmux session. '
                 'Use `update-alternatives --config x-terminal-emulator` to '
                 'configure the terminal.')
        autoparser.add_argument(
            '-i', '--ignore', action='append', metavar='DISK',
            default=[self.block_device(i) for i in settings.IGNORE_DEVICES],
            type=self.block_device,
            help='Ignore the device for auto commands.')
        autoparser.add_argument(
            'action', choices=['start', 'secure-erase'], default='start',
            help='Start docktool or secure erase newly inserted disks')

        notifyparser = self.subparsers.add_parser(
            'auto-notify',
            help='Notify the auto process of the device')
        notifyparser.add_argument(
            'device', metavar='DISK', type=self.block_device,
            help='Device to use for example: /dev/sda')

    def parse_args(self):
        if len(sys.argv) == 2 and not (
                sys.argv[1].startswith('-')
                or sys.argv[1] in self.subparsers.choices):
            sys.argv[1:1] = ['dev']  # imply "dev"
        return super().parse_args()

    def exit(self, status=0, message=None):
        if self.message2:
            print('{prog}: {message}'.format(
                prog=self.prog, message=self.message2),
                file=sys.stderr)
            self.message2 = None
        elif len(sys.argv) == 1:
            try:
                candidates = sorted(set([
                    os.path.realpath(os.path.join('/dev/disk/by-id', i))
                    for i in os.listdir('/dev/disk/by-id')]))
            except FileNotFoundError:
                # ??? no /dev/disk/by-id?
                candidates = ['(no disks found?)']
            message2 = 'suggesting disks:{}'.format(
                '\n  '.join([''] + candidates))
            print('{prog}: {message}'.format(
                prog=self.prog, message=message2),
                file=sys.stderr)

        super().exit(status=0, message=message)

    def block_device(self, name):
        device = Path(name)
        if not device.is_absolute():
            warn('Prepended /dev/ to the block device name')
            return self.block_device('/dev' / device)
        if not device.exists():
            self.message2 = f'{device!r}: not found'
            raise argparse.ArgumentError(f'{device}: not found')
        if not device.is_block_device():
            self.message2 = f'{device!r}: not a block device'
            raise argparse.ArgumentError(f'{device}: not a block device')
        return device


def main():
    parser = DocktoolArgumentParser()
    args = parser.parse_args()

    try:
        if args.command == 'dev':
            dev_menu(args.device, args.action)
        elif args.command == 'serial':
            db_menu(args.serial)
        elif args.command == 'auto':
            AutoServer(args.action, args.gui, args.ignore).start()
        elif args.command == 'auto-notify':
            auto_notify(args.device)
    except KeyboardInterrupt:
        sys.exit(0)
    except SystemExit:
        if getattr(args, 'wait', False):
            input('Press enter to exit')
    except Exception as e:
        print('ERROR: {}\n'.format(e), file=sys.stderr)
        print()
        print(traceback.format_exc(), file=sys.stderr)
        input('Error found, please inform developer. Press enter')
        # #raise e  # (a bit duplicate, no?)
        sys.exit(1)


if __name__ == '__main__':
    main()



from __future__ import print_function

from warnings import warn
import os.path
import re
import subprocess

from .devtype import DEVTYPE_HDD, DEVTYPE_NVME, DEVTYPE_SSD

__author__ = 'osso'


class SmartDataError(Exception):
    pass


class SmartDataParser(object):
    def __init__(self, devname):
        self.devname = devname
        self.rawdata = self._get_smart_data()
        self.information, self.attributes = (
            self._parse_smart_information(self.rawdata))
        self.devtype = self._get_devtype()

    def _get_devtype(self):
        if 'NVMe' in self.rawdata:
            devtype = DEVTYPE_NVME
        elif 'solid state' in self.information.get(
                'Rotation Rate', '').lower():
            devtype = DEVTYPE_SSD
        else:
            devtype = DEVTYPE_HDD
        return devtype

    def _get_smart_data(self):
        smartctl = '/usr/sbin/smartctl'
        try:
            p = subprocess.Popen(
                [smartctl, '-x', self.devname],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = p.communicate()
        except FileNotFoundError as e:
            raise SmartDataError('{} is not installed'.format(
                smartctl)) from e

        # Some devices return junk as part of the vendor name.
        stdout = stdout.decode('utf-8', 'ignore')
        stderr = stderr.decode('utf-8', 'ignore')

        if p.returncode & (
                0x1 |  # "Command line did not parse"
                0x2):  # "Device open failed or ..."
            raise SmartDataError('smartctl error {} on {}: {}'.format(
                p.returncode, self.devname, (
                    stderr + '\n\n' + stdout).strip()))

        return stdout

    @property
    def device_model(self):
        """
        Device Model: Samsung SSD 850 PRO 256GB
        """
        information = self.information
        if 'Model Number' in information:
            return information['Model Number']
        if 'Device Model' in information:
            return information['Device Model']
        if 'Product' in information:
            return information['Product']
        raise ValueError('no modelname found in {}'.format(information))

    @property
    def model_family(self):
        """
        Model Family: Samsung based SSDs
        """
        if 'Model Family' in self.information:
            return self.information['Model Family']
        return None

    @property
    def serial_number(self):
        """
        Serial Number: S1SUNSAF81L482Z
        """
        information = self.information
        key_options = [
            x for x in information.keys() if x.lower() == 'serial number']
        if len(key_options) == 1:
            return information[key_options[0]]
        raise Exception(
            'Serial number could not be determined from info: {0}'.format(
                information))

    @property
    def user_capacity(self):
        """
        User Capacity: 256,060,514,304 bytes [256 GB]
        """
        information = self.information
        if 'User Capacity' in information:
            return information['User Capacity']
        if 'Total NVM Capacity' in information:
            return information['Total NVM Capacity']
        if 'Namespace 1 Size/Capacity' in information:
            return information['Namespace 1 Size/Capacity']
        raise Exception(
            'User capacity could not be determined from info: {0}'.format(
                information))

    @property
    def firmware_version(self):
        """
        Firmware Version: EXM01B6Q
        """
        return self.information.get('Firmware Version', 'unknown')

    @property
    def ata_version(self):
        """
        ATA Version is: ACS-2, ATA8-ACS T13/1699-D revision 4c
        """
        return self.information.get('ATA Version is', 'unknown')

    @property
    def smart_status(self):
        """
        SMART support is: Available - device has SMART capability.
        SMART support is: Enabled  <-- 2nd one is kept!
        SMART overall-health self-assessment test result: PASSED
        """
        if self.information.get('SMART support is') != 'Enabled':
            print('WARNING: SMART support not enabled? {!r} + {!r}'.format(
                self.information.get('SMART support is'),
                self.information.get('SMART overall-health')))
            return 'MISSING'
        elif not self.information.get('SMART overall-health'):
            print('WARNING: SMART not passed?')
            return 'Enabled'
        return self.information.get('SMART overall-health')

    @property
    def sector_size(self):
        """
        Sector Size: 512 bytes logical/physical
        """
        match = re.match(r'([0-9]*) ', self.information.get('Sector Size', ''))
        if match:
            return int(match.groups()[0])
        # Default to 512 (bytes)
        return 512

    def _get_value(self, search_str, column_name):
        data = [
            x for x in self.attributes
            if x.get('attribute_name', '').lower() == search_str]

        if data:
            value = data[0].get(column_name, None)
            try:
                value = int(value)
            except ValueError:
                pass
            return value
        return None

    @property
    def wear_health_percent(self):
        """
        Returns estimated health of device (from 100 down to 0 (or below))
        """
        # This is a percentage (the value, that is), which goes down from 100%
        # to 0% (= no health); taken from the obscurely named
        # "wear_leveling_count".
        keys = ('media_wearout_indicator', 'wear_leveling_count')
        for key in keys:
            wear = self._get_value(key, 'value')
            if wear is not None:
                break
        else:
            if self.devtype == DEVTYPE_HDD:
                # Spinners have no wearout indication.
                wear = 100  # 100% ok == "OK".. unless we have any other info..
            else:
                raise ValueError(
                    'no wear attribute found for {} (type {!r}), '
                    'please file bug'.format(self.device_model, self.devtype))
        return int(wear)

    @property
    def power_on_hours(self):
        return self._get_value('power_on_hours', 'raw_value')

    @property
    def lbas_read(self):
        return self._get_value('total_lbas_read', 'raw_value')

    @property
    def lbas_written(self):
        return self._get_value('total_lbas_written', 'raw_value')

    @property
    def reallocated_sector_ct(self):
        return self._get_value('reallocated_sector_ct', 'raw_value')

    @property
    def reallocated_event_count(self):
        return self._get_value('reallocated_event_count', 'raw_value')

    @property
    def current_pending_sector(self):
        return self._get_value('current_pending_sector', 'raw_value')

    @property
    def offline_uncorrectable(self):
        return self._get_value('offline_uncorrectable', 'raw_value')

    def _parse_smart_information(self, rawdata):
        it = iter(rawdata.split('\n'))

        information = {}
        for line in it:
            if 'START OF INFORMATION SECTION' in line:
                break
        # === START OF INFORMATION SECTION ===
        # Model Family:     Samsung based SSDs
        # Device Model:     Samsung SSD 840 EVO 120GB
        for line in it:
            line = line.strip()
            if not line:
                break
            if ':' not in line:
                warn('smartctl line without KEY/VALUE: {}'.format(line))
            else:
                key, value = line.split(':', 1)
                key = key.strip()
                if key != 'SMART support is':
                    assert key not in information, ((key, value), information)
                value = value.strip()
                information[key] = value
        assert information, 'No info found in smart data:\n{}'.format(
            rawdata)

        attributes = []
        for line in it:
            # === START OF READ SMART DATA SECTION ===
            # SMART overall-health self-assessment test result: PASSED
            # SMART overall-health self-assessment test result: FAILED!
            if line.startswith('SMART overall-health self-assessment') and (
                    ':' in line):
                information['SMART overall-health'] = (
                    line.split(':', 1)[1].strip())
            elif line.startswith('SMART Health Status:'):
                # SAS disk, Seagate ST10000NM0226
                information['SMART overall-health'] = (
                    line.split(':', 1)[1].strip())

            # ID# ATTRIBUTE_NAME  FLAGS  VALUE ...
            elif 'ID# ATTRIBUTE_NAME' in line:
                attribute_headers = [
                    i.lower() for i in re.split(r' +', line)]
                break
        # ID# ATTRIBUTE_NAME          FLAGS    VALUE WORST THRESH FAIL RAW_VAL
        #   5 Reallocated_Sector_Ct   PO--CK   100   100   010    -    0
        #   9 Power_On_Hours          -O--CK   080   080   000    -    99644
        #  12 Power_Cycle_Count       -O--CK   099   099   000    -    44
        # 177 Wear_Leveling_Count     PO--C-   082   082   000    -    216
        for line in it:
            line = line.strip()
            split = re.split(r' +', line)
            if '|' in split[0]:  # ||||| pointing to the FLAGS
                break
            values = {}
            for nr, key in enumerate(attribute_headers):
                values[key] = split[nr]
            attributes.append(values)

        if not attributes:
            warn('smartctl: No attributes found')

        return information, attributes


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 2:
        print('Usage {} device'.format(
            os.path.basename(sys.argv[0])), file=sys.stderr)
        sys.exit(1)

    # ID# ATTRIBUTE_NAME          FLAGS  VALUE WORST THRESH FAIL RAW_VALUE
    #   5 Reallocated_Sector_Ct   PO--CK 100   100   010    -    0
    #   9 Power_On_Hours          -O--CK 084   084   000    -    76546
    #  12 Power_Cycle_Count       -O--CK 099   099   000    -    22
    # 177 Wear_Leveling_Count     PO--C- 076   076   000    -    1452
    # 179 Used_Rsvd_Blk_Cnt_Tot   PO--C- 100   100   010    -    0
    # 181 Program_Fail_Cnt_Total  -O--CK 100   100   010    -    0
    # 182 Erase_Fail_Count_Total  -O--CK 100   100   010    -    0
    # 183 Runtime_Bad_Block       PO--C- 100   100   010    -    0
    # 187 Uncorrectable_Error_Cnt -O--CK 100   100   000    -    0
    # 190 Airflow_Temperature_Cel -O--CK 069   037   000    -    31
    # 195 ECC_Error_Rate          -O-RC- 200   200   000    -    0
    # 199 CRC_Error_Count         -OSRCK 100   100   000    -    0
    # 235 POR_Recovery_Count      -O--C- 099   099   000    -    8
    # 241 Total_LBAs_Written      -O--CK 099   099   000    -    167448209929

    parser = SmartDataParser(sys.argv[1])
    for prop in (
            'device_model',
            'model_family',
            'serial_number',
            'user_capacity',
            'firmware_version',
            'ata_version',
            'smart_status',
            'sector_size',
            'lbas_written',
            'lbas_read',
            'power_on_hours',
            'wear_health_percent',
            'reallocated_sector_ct',
            'reallocated_event_count',
            'current_pending_sector',
            'offline_uncorrectable'):
        val = getattr(parser, prop)
        str_val = str(val) if val is not None else '(N/A)'
        print('{:24}: {}'.format(prop, str_val))



import os
import re
import subprocess
import time

from . import settings
from .devtype import DEVTYPE_HDD, DEVTYPE_NVME, DEVTYPE_SSD
from .nvmedata import NvmeDataError, NvmeDataParser
from .smartdata import SmartDataError, SmartDataParser
from .hdparm import HdparmParser

__author__ = 'osso'


class UnmatchedStorageDevice(Exception):
    pass


class BaseStorageDevice(object):
    @classmethod
    def from_devname(cls, devname):
        for class_ in (
                NvmeStorageDevice,
                SgSanitizeStorageDevice,
                FallbackStorageDevice,
                ):
            try:
                devtype, hwdata = class_._get_hwdata(devname)
            except UnmatchedStorageDevice as e:
                last_exception = e
            else:
                return class_(devname, devtype, hwdata)
        raise last_exception

    @classmethod
    def _get_hwdata(cls, devname):
        raise NotImplementedError()

    def __init__(self, devname, devtype, hwdata):
        self.devname = devname
        self.devtype = devtype
        self.hwdata = hwdata
        self.status = 'Initialized'
        self.port = self._get_port()
        self.docktool_bay_nr = self._get_docktool_bay_nr(self.port)

    def flush(self):
        self.devtype, self.hwdata = self._get_hwdata(self.devname)

    def get_device_model(self):
        # "Samsung SSD 850 PRO 256GB" or ValueError
        return self.hwdata.device_model

    def get_model_family(self):
        # "Samsung based SSDs" or None
        return self.hwdata.model_family

    def get_serial_number(self):
        # "S1SUNSAF81L482Z" or ValueError
        return self.hwdata.serial_number

    def get_user_capacity(self):
        # "256,060,514,304 bytes [256 GB]" or None
        return self.hwdata.user_capacity

    def get_firmware_version(self):
        # "EXM01B6Q" or "unknown"
        return self.hwdata.firmware_version

    def get_ata_version(self):
        # "ACS-2, ATA8-ACS T13/1699-D revision 4c" or "unknown"
        return self.hwdata.ata_version

    def get_smart_status(self):
        # "PASSED" or "OK" or "FAILED!" or "Enabled" or "MISSING"
        return self.hwdata.smart_status

    def is_sas(self):
        return 'sas' in self.hwdata.information.get(
            'Transport protocol', '').lower()

    def is_ssd(self):
        return (self.devtype in (DEVTYPE_NVME, DEVTYPE_SSD))

    def _get_docktool_bay_nr(self, port):
        return settings.ATA_TO_BAY_NR.get(port, None)

    def _get_port(self):
        realpath = os.path.realpath(self.devname)
        bypath = '/dev/disk/by-path/'
        found_symlinks = []
        for symlink in os.listdir(bypath):
            if os.path.realpath(os.path.join(bypath, symlink)) == realpath:
                found_symlinks.append(symlink)
        port = None
        if found_symlinks:
            assert len(found_symlinks) == 1
            for regex in [
                    # 'pci-0000:00:1f.2-ata-1'
                    r'-(ata-\d+)',
                    # 'pci-0000:17:00.0-nvme-1'
                    r'(^pci-.*nvme-\d+)',
                    # 'pci-0000:19:00.0-sas-e..f-phy5-lun-0'
                    r'-(phy\d+)',
                    # 'pci-0000:00:14.0-usb-0:13.2:1.0-scsi-0:0:0:0'
                    r'-(usb-\d+)']:
                match = re.search(regex, found_symlinks[0])
                if match:
                    port = match.groups()[0]
                    break

        return port or self.devname

    def get_erase_methods(self):
        return [self.wipe_using_manual_methods]

    def wipe_using_manual_methods(self):
        """
        Wipe using modified urandom stream + zerofill.

        Slow.
        """
        if not settings.DEBUG:
            # secure erase by using /dev/urandom
            subprocess.check_output([
                '/usr/local/bin/run-lessrandom', self.devname])
            subprocess.check_output([
               '/usr/local/bin/run-zero-disk', self.devname])
            self.status = 'Secure erased'
        else:
            print('WARNING: Skipping manual erase because DEBUG')

        return True, None
    wipe_using_manual_methods.name = 'manual, slow'
    wipe_using_manual_methods.type = 'manual'  # manual = slow

    def quick_erase(self):
        if not settings.DEBUG:
            # Quick erase by zeroing begin and end of disk:
            # - 160M at the start
            startbs, startcount = (1024 * 1024), 160
            # - 104M at the end
            endbs, endcount = 512, 204800

            with open(self.devname, 'rb') as fp:
                size = fp.seek(0, 2)

            subprocess.check_output([
                'dd',
                'if=/dev/zero',
                'of={}'.format(self.devname),
                'bs={}'.format(startbs),
                'seek=0',
                'count={}'.format(startcount),
                'conv=fsync'])
            subprocess.check_output([
                'dd',
                'if=/dev/zero',
                'of={}'.format(self.devname),
                'bs={}'.format(endbs),
                'count={}'.format(endcount),
                'seek={}'.format((size // endbs) - endcount),
                'conv=fsync'])
        else:
            print('WARNING: Skipping quick erase because DEBUG')

        self.status = 'Quick erased'

    def shutdown_disk(self):
        print('\nSpinning down/ejecting disk, please wait...\n')
        if settings.DEBUG:
            print('WARNING: Skipping spindown/eject because DEBUG')
        elif self.is_sas():
            subprocess.check_output([
                '/usr/bin/sdparm', '--readonly', '--command=stop',
                self.devname])
        elif self.devtype == DEVTYPE_NVME:
            print('WARNING: No idea how to eject and NVMe dev; needed?')
        else:
            assert self.devtype in (DEVTYPE_HDD, DEVTYPE_SSD)
            # FIXME: what if we don't want to spin it down? because we're not
            # ejecting it
            subprocess.check_output(['/sbin/hdparm', '-Y', self.devname])
        self.status = 'Shutdown'


class GenericStorageDeviceBase(BaseStorageDevice):
    @classmethod
    def _get_hwdata(cls, devname):
        try:
            hwdata = SmartDataParser(devname)
        except SmartDataError as e:
            raise UnmatchedStorageDevice(e) from e
        if hwdata.devtype == DEVTYPE_NVME:
            raise NotImplementedError(
                'programming error; we should be in NvmeStorageDevice')
        return hwdata.devtype, hwdata


class SgSanitizeStorageDevice(GenericStorageDeviceBase):
    @classmethod
    def _get_hwdata(cls, devname):
        devtype, hwdata = super()._get_hwdata(devname)

        # Is this as SAS disk, then assume hdparm security yields None
        # and we might get crypto erase support.
        if 'sas' in hwdata.information.get('Transport protocol', '').lower():
            # Assert that a SAS disk does not do hdparm stuff.
            hdparm_sec_info = HdparmParser(devname).get_security_info()
            assert hdparm_sec_info is None, (
                hdparm_sec_info, hwdata.information)

            # Try sg_opcodes.
            opcodes = cls._get_sg_security_info(devname)
            if opcodes['sanitize_crypto_erase']:
                return devtype, hwdata

        raise UnmatchedStorageDevice()

    def get_erase_methods(self):
        return [
            self.wipe_using_sanitize_crypto_erase,
            self.wipe_using_manual_methods]

    def wipe_using_sanitize_crypto_erase(self):
        # /usr/bin/sg_sanitize --quick --wait --verbose --crypto devname
        if not settings.DEBUG:
            if not (
                    self._get_sg_security_info(self.devname)
                    ['sanitize_crypto_erase']):
                return False, 'no sanitize_crypto_erase available'
            subprocess.check_call([
                '/usr/bin/sg_sanitize', '--quick', '--wait', '--verbose',
                '--crypto', self.devname])
            return True, None
        else:
            print('WARNING: Skipping crypto erase erase because DEBUG')
            return True, None
    wipe_using_sanitize_crypto_erase.name = 'sanitize crypto erase'
    wipe_using_sanitize_crypto_erase.type = 'builtin'  # fast!

    @classmethod
    def _get_sg_security_info(cls, devname):
        opcodes = cls._get_sg_opcodes(devname)
        ret = {
            # 'sanitize_overwrite': False,
            'sanitize_crypto_erase': False,
        }
        if (0x48, '1') in opcodes:
            # This, is no fun. I tested one Seagate 10TB disk, and it did
            # 175MB/s for the first 3.78% (in 36 minutes). That translates to
            # 16 hours for the total disk.
            # # sg_requests /dev/sda
            # > Decode parameter data as sense data:
            # > Fixed format, current; Sense key: Not Ready
            # > Additional sense: Logical unit not ready, sanitize in progress
            # >   Progress indication: 3.78%
            # Additionally, it might even have stalled now, as the status
            # update from sg_requests totally hung at t0+4hours.
            # #ret['sanitize_overwrite'] = True
            pass
        if (0x48, '3') in opcodes:
            ret['sanitize_crypto_erase'] = True
        return ret

    @classmethod
    def _get_sg_opcodes(cls, devname):
        # > # sg_opcodes --compact /dev/sda
        # >   SEAGATE   ST10000NM0226     KT01
        # >   Peripheral device type: disk
        # >
        # > Opcode,sa  Name
        # >   (hex)
        # > ---------------------------------------
        # >  00        Test Unit Ready
        # >  01        Rezero Unit
        # >  03        Request Sense
        # > ...
        # >  48,1      Sanitize, overwrite
        # >  48,3      Sanitize, cryptographic erase
        # >  48,1f     Sanitize, exit failure mode
        opcodes = (
            subprocess.check_output(
                ['/usr/bin/sg_opcodes', '--compact', devname])
            .decode('utf-8'))
        it = iter(opcodes.strip().split('\n'))
        for line in it:
            if line.startswith('----'):
                break
        ret = {}
        for line in it:
            opcode, title = line.strip().split(None, 1)
            opcodes = opcode.split(',')
            opcodes[0] = int(opcodes[0], 16)
            assert tuple(opcodes) not in ret, (ret, opcodes)
            ret[tuple(opcodes)] = title
        return ret


class FallbackStorageDevice(GenericStorageDeviceBase):
    def get_erase_methods(self):
        if self._can_secure_erase()[0]:
            return [
                self.wipe_using_security_erase, self.wipe_using_manual_methods]
        return [self.wipe_using_manual_methods]

    def _can_secure_erase(self):
        security_info = HdparmParser(self.devname).get_security_info()
        if security_info is None:
            return False, 'no hdparm security, no sg_opcode security'

        for x in ['has_security', 'is_enabled', 'is_locked', 'is_frozen']:
            if x not in security_info:
                return False, '{} not found in security_info: {}'.format(
                    x, security_info)

        if not security_info['has_security']:
            return False, 'Block device secure erase seems not to be supported'

        if security_info['is_enabled']:
            assert False, (
                'Block device password for secure erase is already set')

        if security_info['is_locked']:
            assert False, (
                'Block device is locked (try replugging the disk or '
                'hibernating system)')

        if security_info['is_frozen']:
            assert False, (
                'Block device is frozen (try replugging the disk or '
                'hibernating system)')

        if security_info.get('is_past_retry_count'):
            assert False, (
                'Block device is past password retry count (time to look up '
                'a master password?)')

        return True, None

    def wipe_using_security_erase(self):
        # $1 = /dev/sdc
        # secret=`pwgen -s 32 1`
        # hdparm --user-master u --security-set-pass $secret $1
        # hdparm --user-master u --security-erase $secret $1
        if not settings.DEBUG:
            can_erase, error = self._can_secure_erase()
            if not can_erase:
                return False, error
            return self._wipe_using_hdparm_security_erase()
        else:
            print('WARNING: Skipping secure erase because DEBUG')
            return True, None
    wipe_using_security_erase.name = 'secure SSD/HDD erase'
    wipe_using_security_erase.type = 'builtin'  # slow? fast?

    def _wipe_using_hdparm_security_erase(self):
        # Create secret
        # XXX: this is problematic if the system dies before we've called
        # security-erase: afterwards the disk sometimes stays locked. It'd
        # be nicer if we wiped with a pre-existing pw?
        secret = subprocess.check_output([
            '/usr/bin/pwgen', '-s', '32', '1'])[0:32].decode('utf-8')

        # Set the password. This should set is_enabled ('enabled')
        # in hdparm security info.
        print('WARNING: The hdparm --security-set-pass sometimes says: ')
        print('  "SG_IO: bad/missing sense data, sb[]:  70 00 01 00 ..."')
        print('This might be ignorable. Please inform developer.')
        print(
            'NOTICE: Using secret {!r}.. you may need this to unbrick!'
            .format(secret))

        print('> --security-set-pass')
        subprocess.check_output([
            '/sbin/hdparm', '--user-master', 'u',
            '--security-set-pass', secret, self.devname])

        # Check hdparm security info. Only perform secure BLKDEV
        # erase if security is indeed enabedl.
        security_info = HdparmParser(self.devname).get_security_info()
        print('> {}'.format(security_info))
        if not security_info.get('is_enabled'):
            return False, (
                'Block device password could not be set, '
                'security info: {}'.format(security_info))

        for i in range(10):
            # Try secure erase.
            if security_info.get('has_enhanced_erase', False):
                print('> --security-erase-enhanced')
                subprocess.check_output([
                    '/sbin/hdparm', '--user-master', 'u',
                    '--security-erase-enhanced', secret, self.devname])
            else:
                print('> --security-erase')
                subprocess.check_output([
                    '/sbin/hdparm', '--user-master', 'u',
                    '--security-erase', secret, self.devname])
            # Check that hdparm security is now unset.
            security_info = HdparmParser(self.devname).get_security_info()
            print('> {}'.format(security_info))
            if not security_info.get('is_enabled', True):
                break
            # This could be a problem.. retry this a few times..
            print(
                'WARNING: Security erase/unset failed (password {!r});'
                'retrying in 3..'.format(secret))
            time.sleep(3)
        else:
            return False, (
                'Block device password could not be UNset, '
                'security info: {}, secret: {!r}'.format(
                    security_info, secret))

        return True, None


class NvmeStorageDevice(BaseStorageDevice):
    @classmethod
    def _get_hwdata(cls, devname):
        try:
            hwdata = NvmeDataParser(devname)
        except NvmeDataError as e:
            raise UnmatchedStorageDevice(e) from e
        assert hwdata.devtype == DEVTYPE_NVME, hwdata.devtype
        return hwdata.devtype, hwdata

    def get_erase_methods(self):
        methods = []
        flags = self._get_secure_erase_options()

        # We prefer:
        # - nvme-sanitize (wipes data AND cache)
        #     over
        # - nvme-format (wipes data).
        # We prefer:
        # - crypto-erase
        #     over
        # - no-secure-erase

        # However: the block erase is a lot slower than crypto erase..
        # What to do, what to do?
        if flags['can_crypto_sanitize']:
            methods.append(self.wipe_using_crypto_sanitize)  # fast
        if flags['can_block_sanitize']:
            methods.append(self.wipe_using_block_sanitize)  # slow
        if flags['can_crypto_erase']:
            methods.append(self.wipe_using_crypto_erase)  # fast
        if flags['can_format']:
            methods.append(self.wipe_using_block_erase)  # fast
        methods.append(self.wipe_using_manual_methods)
        return methods

    def wipe_using_block_erase(self):
        if not settings.DEBUG:
            subprocess.check_output(
                ['/usr/sbin/nvme', 'format', self.devname, '--ses=1'])
        else:
            print('WARNING: Skipping secure block erase because DEBUG')
        return True, None
    wipe_using_block_erase.name = 'nvme format (poor)'
    wipe_using_block_erase.type = 'builtin'  # ???

    def wipe_using_block_sanitize(self):
        if not settings.DEBUG:
            subprocess.check_output(
                ['/usr/sbin/nvme', 'sanitize', self.devname, '--sanact=0x2'])
        else:
            print('WARNING: Skipping secure block sanitize because DEBUG')
        return True, None
    wipe_using_block_sanitize.name = 'nvme block sanitize (better)'
    wipe_using_block_sanitize.type = 'builtin'  # ???

    def wipe_using_crypto_erase(self):
        if not settings.DEBUG:
            subprocess.check_output(
                ['/usr/sbin/nvme', 'format', self.devname, '--ses=2'])
        else:
            print('WARNING: Skipping secure crypto erase because DEBUG')
        return True, None
    wipe_using_crypto_erase.name = 'nvme crypto erase (better)'
    wipe_using_crypto_erase.type = 'builtin'  # ???

    def wipe_using_crypto_sanitize(self):
        if not settings.DEBUG:
            subprocess.check_output(
                ['/usr/sbin/nvme', 'sanitize', self.devname, '--sanact=0x4'])
        else:
            print('WARNING: Skipping secure crypto sanitize because DEBUG')
        return True, None
    wipe_using_crypto_sanitize.name = 'nvme crypto sanitize (best)'
    wipe_using_crypto_sanitize.type = 'builtin'  # ???

    def _get_secure_erase_options(self):
        if self.devtype != DEVTYPE_NVME:
            return False, 'Disk was not recognized as NVMe Device'

        # > Secure erase is faster, but not as complete. Data on a drive that
        # > has been securely erased or sanitized cannot be recovered. If power
        # > is interrupted during a secure erase, secure erase may continue as
        # > soon as power is restored or the user may need to issue another
        # > secure erase command when power is restored. Note that SSD
        # > manufacturers do not follow a universal command to sanitize drives.

        # "Format NVM Supported"
        can_format = False
        if self.hwdata.information['oacs:1']:  # 2nd bit of oacs
            assert bool(self.hwdata.information['oacs'] & 0x2)  # same info
            can_format = True
            # -> nvme format /dev/nvme0 --ses=1  # nvme format/wipe

        # "Crypto Erase Supported as part of Secure Erase"
        can_crypto_erase = False
        if can_format and self.hwdata.information['fna:2']:
            assert bool(self.hwdata.information['fna'] & 0x4)  # same info
            can_crypto_erase = True
            # -> nvme format /dev/nvme0 --ses=2  # nvme crypto format

        # "Block Erase Sanitize Operation (Not) Supported"
        can_block_sanitize = False
        # "Crypto Erase Sanitize Operation (Not) Supported"
        can_crypto_sanitize = False
        if self.hwdata.information['sanicap']:
            if self.hwdata.information['sanicap:1']:
                assert bool(self.hwdata.information['sanicap'] & 0x2)  # same
                can_block_sanitize = True
                # # 010b - Start a Block Erase sanitize operation
                # # 011b - Start an Overwrite sanitize operation
                # # 100b - Start a Crypto Erase sanitize operation
                # -> nvme sanitize /dev/nvme0 --sanact=0x2  # block sanitize
            if self.hwdata.information['sanicap:0']:
                assert bool(self.hwdata.information['sanicap'] & 0x1)  # same
                can_crypto_sanitize = True
                # # 010b - Start a Block Erase sanitize operation
                # # 011b - Start an Overwrite sanitize operation
                # # 100b - Start a Crypto Erase sanitize operation
                # -> nvme sanitize /dev/nvme0 --sanact=0x4  # crypto sanitize

        # can_crypto_erase = True && can_crypto_sanitize = False
        # --> means quick erase (but not best)
        # --> or slow erase (crypto sanitize)
        return {
            'can_crypto_sanitize': can_crypto_sanitize,
            'can_crypto_erase': can_crypto_erase,
            'can_block_sanitize': can_block_sanitize,
            'can_format': can_format,
        }



from time import time
import base64
import json

import requests

try:
    from urllib.parse import urljoin, urlparse, parse_qs  # py3
except ImportError:
    from urlparse import urljoin, urlparse, parse_qs  # py2

from . import settings


def base64_decode(value):
    return base64.b64decode(value + '=' * (-len(value) % 4))


def get_expiry(jwt_token):
    payload = json.loads(
        base64_decode(jwt_token.split('.')[1]).decode('ascii'))
    exp = payload.get('exp', 0)
    return int(exp - time())


def is_expired(jwt_token):
    return ((get_expiry(jwt_token) - 10) < 0)


class JWTTokenProvider(object):
    authserver_token_path = '/api/user/token/'

    def __init__(self, login_url_or_auth_scheme_url):
        # Either SERVER/api/auth/scheme/ or SERVER/api-auth/login/.
        self._first_url = login_url_or_auth_scheme_url
        self.client = requests.session()

    def get_authorize_url(self):
        res = self.client.get(self._first_url, allow_redirects=False)
        if 300 <= res.status_code < 400:
            # https://AUTHSERVER/openid/authorize?scope=openid+profile&'
            #   redirect_uri=REDIR_URI&response_type=code&client_id=CLIENT_CODE
            location = res.headers.get('Location')
            if 'response_type=code' in location:
                return location
        elif res.status_code == 200:
            # {"data":{"authorizeUrl":"LOCATION_AS_SEEN_ABOVE"},
            # "type":"oidc","name":"oidc"}
            data = res.json()
            if data['type'] == 'oidc':
                return data['data']['authorizeUrl']
        raise ValueError('Unknown scheme {} -> {} {!r}'.format(
            self._first_url, res.status_code, res.text))

    def get_token_url(self):
        # We don't seem to need scope. Just use the client_id.
        url = self.get_authorize_url()
        parsed = urlparse(url)
        qs = parse_qs(parsed.query)
        client_id = qs.get('client_id')
        assert len(client_id) == 1 and client_id[0], url
        return '{}?client_id={}'.format(
            # urljoin chops off the original query_string.
            urljoin(parsed.geturl(), self.authserver_token_path), client_id[0])

    def get_jwt_token(self, auth_token=None):
        url = self.get_token_url()

        if auth_token:
            headers = {'Authorization': 'Token {}'.format(auth_token)}
            kwargs = {'headers': headers}
        else:
            raise NotImplementedError()

        res = self.client.get(url, **kwargs)
        if res.status_code != 200:
            raise ValueError('Bad token response {} -> {} {!r}'.format(
                url, res.status_code, res.text))
        js = res.json()
        assert js.get('token_type') == 'Bearer', (url, js)
        return js['id_token']


def get_jwt_token(scheme_url, auth_token):
    prov = JWTTokenProvider(scheme_url)
    return prov.get_jwt_token(auth_token=auth_token)


def main():
    prov = JWTTokenProvider(settings.DASHBOARD_SCHEME_URL)
    jwt = prov.get_jwt_token(auth_token=settings.DASHBOARD_AUTH_TOKEN)
    print('Got JWT')
    print(jwt)
    print('Expires in {}s'.format(get_expiry(jwt)))


if __name__ == '__main__':
    main()



from base64 import b16encode
from hashlib import md5
import random
import time

__author__ = 'osso'


def hexdigest(data):
    return b16encode(data).decode('ascii').lower()


class SpuriousReadError(OSError):
    pass


class DiskEmptySampler(object):
    """
    Divide the disk in 'sample_count' regions
    and sample (md5sum) 'sample_size' bytes randomly per region
    """
    def __init__(self, devname, sample_count=100, sample_size=(1024 * 1024)):
        self.devname = devname

        # No divide by zero errors
        if sample_count == 0:
            raise Exception('Sample size should not be zero')

        self.sample_count = sample_count
        self.sample_size = sample_size

        self.sample_regions = None
        self.sample_runs = []

        self.expected_hash = hexdigest(md5(b'\x00' * sample_size).digest())

    def sample(self):
        attempts = 3
        for attempt in range(1, attempts + 1):
            try:
                samples = self._sample_disk()
            except SpuriousReadError:
                if attempt == attempts:
                    raise
                print('WARNING: Read error. Possibly too fast after disk wipe')
                time.sleep(5)
                print('Retrying...')
            else:
                self.sample_runs.append(samples)
                break

    def is_zero(self):
        assert self.sample_runs, repr(self.sample_runs)
        samples = self.sample_runs[-1]
        assert samples, repr(samples)

        for sample_nr, md5_sum in enumerate(samples):
            if hexdigest(md5_sum) != self.expected_hash:
                print('Digests are unexpected: {} != {}'.format(
                    hexdigest(md5_sum), self.expected_hash))
                print('for offset: {} (sample_size: {})'.format(
                    self.sample_regions[sample_nr], self.sample_size))
                return False

        return True

    def is_different(self):
        assert self.sample_runs and len(self.sample_runs) >= 2, repr(
            self.samples)
        samples_1 = self.sample_runs[-2]
        samples_2 = self.sample_runs[-1]

        both_samples = zip(samples_1, samples_2)
        for sample_nr, (md5_1, md5_2) in enumerate(both_samples):
            if md5_1 == md5_2:
                print('Digests are equal between runs: {}'.format(
                    hexdigest(md5_1)))
                print('for offset: {} (sample_size: {})'.format(
                    self.sample_regions[sample_nr], self.sample_size))
                return False
        return True

    def _sample_disk(self):
        with open(self.devname, 'rb') as fp:
            if not self.sample_regions:
                self.sample_regions = self._generate_sample_regions(
                    fp, self.sample_count, self.sample_size)

            samples = []

            # Loop over the sample_regions (offsets) and store the md5 of that
            # region.
            assert self.sample_regions
            for sample_nr, offset in enumerate(self.sample_regions):
                # Jump there, and fetch block.
                fp.seek(offset)
                try:
                    data = fp.read(self.sample_size)
                except OSError as e:
                    # On NVMe drives, we have seen that a read() here
                    # fails immediately after an nvme sanitize
                    # operation. The SpuriousReadError allows us to back
                    # off and retry after a little sleep.
                    errno = e.args[0]
                    if errno == 5:  # EIO
                        raise SpuriousReadError() from e
                    raise

                # Create md5 and store for later inspection.
                samples.append(md5(data).digest())

        return samples

    @staticmethod
    def _generate_sample_regions(fp, sample_count, sample_size):
        regions = []

        # Get the size of disk (e.g. 256GB).
        size = fp.seek(0, 2)

        # Divide the disk into sample_regions (e.g. 256GB/2000 = ~131MB).
        sample_region_size = size // sample_count

        # Check if the regions are large enough to get a sample from.
        if sample_size > sample_region_size:
            raise ValueError('Sample size is too large: {} > {}/{}'.format(
                sample_size, size, sample_count))

        # Loop over the sample_regions and select a random sample inside
        # that region:
        # - 0: first 131MB: choose a random sample_size block there
        # - 1: second 131MB: choose a random sample_size block there
        # ..etc..
        for sample_nr in range(sample_count):
            # Calculate begin and end of region based on sample_region
            # size and pick a random offset in it:
            region_start = sample_nr * sample_region_size
            region_end_ex = min((sample_nr + 1) * sample_region_size, size)
            offset = random.randint(
                region_start, region_end_ex - sample_size)
            regions.append(offset)

        return regions



import subprocess
import unittest

__author__ = 'osso'


class bool_dict(dict):
    def __str__(self):
        return '{{{}}}'.format(', '.join(
            k if v else '!{}'.format(k)
            for k, v in sorted(self.items())))


class HdparmParser(object):
    RENAME_MAP = {
        # [...] a counter in the hard disk electronics allows only five
        # attempts to enter the user password. Otherwise, the device
        # remains locked in SEC4 state. Any further attempt at unlocking
        # is only possible after a power cycle or hardware reset.
        'expired: security count': 'is_past_retry_count',
        'supported: enhanced erase': 'has_enhanced_erase',
        # The olden flags, renamed to make more sense.
        'supported': 'has_security',
        'enabled': 'is_enabled',
        'frozen': 'is_frozen',
        'locked': 'is_locked',
    }

    def __init__(self, devname):
        self._rawdata = subprocess.check_output(
            ['/sbin/hdparm', '-I', devname]).decode('utf-8')

    def get_security_info(self):
        if 'Security' not in self._rawdata:
            return None

        lines = self._rawdata.split('\n')
        security_info = {}
        do_collect = False
        for line in lines:
            if do_collect:
                if not line.startswith('\t'):
                    do_collect = False
                else:
                    splitted = line.split('\t')
                    if (len(splitted) == 3 and
                            'master password' not in splitted[1].lower()):
                        security_info[splitted[2]] = (
                            splitted[1].lower() != 'not')
            elif line.startswith('Security'):
                do_collect = True

        return bool_dict(
            (self.RENAME_MAP.get(k, k), v) for k, v in security_info.items())


class HdparmParserTestCase(unittest.TestCase):
    def test_values(self):
        class TestHdparmParser(HdparmParser):
            def __init__(self, rawdata):
                self._rawdata = rawdata

        input = '\n'.join([
            '\t   *\tREAD BUFFER DMA command'
            '\t   *\tData Set Management TRIM supported (limit 8 blocks)',
            'Security: ',
            '\tMaster password revision code = 65534',
            '\t\tsupported',
            '\tnot\tenabled',
            '\tnot\tlocked',
            '\t\tfrozen',
            '\tnot\texpired: security count',
            '\t\tsupported: enhanced erase',
            ('\t2min for SECURITY ERASE UNIT. '
             '8min for ENHANCED SECURITY ERASE UNIT.'),
            'Logical Unit WWN Device Identifier: 50025388a01c561a',
            '\tNAA\t\t: 5',
            '\tIEEE OUI\t: 002538',
        ])
        hdparamparser = TestHdparmParser(input)
        security_info = hdparamparser.get_security_info()
        self.assertEqual(security_info, {
            'is_enabled': False,
            'is_past_retry_count': False,
            'is_frozen': True,
            'is_locked': False,
            'has_security': True,
            'has_enhanced_erase': True,
        })
        self.assertEqual(str(security_info), (
            '{has_enhanced_erase, has_security, '
            '!is_enabled, is_frozen, !is_locked, !is_past_retry_count}'))



import json

import requests

try:
    from urllib.parse import urljoin  # py3
except ImportError:
    from urlparse import urljoin  # py2

from . import settings
from .jwt_token_provider import get_jwt_token, is_expired

__author__ = 'osso'


class InventoryRESTClient(object):
    def __init__(self, base_url):
        self.base_url = urljoin(base_url, '/api/')
        self.hdd_uid = None
        self.__jwt_token = None

    @property
    def auth_header(self):
        if self.__jwt_token:
            if is_expired(self.__jwt_token):
                self.__jwt_token = None
        if not self.__jwt_token:
            self.__jwt_token = get_jwt_token(
                settings.DASHBOARD_SCHEME_URL,
                settings.DASHBOARD_AUTH_TOKEN)
        return {'Authorization': 'JWT {0}'.format(self.__jwt_token)}

    def __post_json(self, url, data):
        headers = {'content-type': 'application/json'}
        headers.update(self.auth_header)
        res = requests.post(
            urljoin(self.base_url, url), data=json.dumps(data),
            headers=headers)
        return res

    def __patch_json(self, url, data):
        headers = {'content-type': 'application/json'}
        headers.update(self.auth_header)
        res = requests.patch(
            urljoin(self.base_url, url),
            data=json.dumps(data),
            headers=headers)
        return res

    def get(self, url):
        return self.__get(url)

    def __get(self, url):
        res = requests.get(
            urljoin(self.base_url, url),
            headers=self.auth_header)
        return res

    def __generic_patch(self, url, data):
        res = self.__patch_json(url, data)
        if res.status_code != 200:
            raise Exception(
                'Status code {}, message {}, when trying to PATCH {}'.format(
                    res.status_code, res.text, data))
        return res.json()

    def __generic_post(self, url, data):
        res = self.__post_json(url, data)
        if res.status_code not in (200, 201):
            raise Exception(
                'Status code {}, message {}, when trying to POST {}'.format(
                    res.status_code, res.text, data))
        return res.json()

    def get_last_owner(self, asset_id):
        res = self.__get(
            'assets/asset_owner/?asset={0}&order_by=-date&limit=1'.format(
                asset_id)
        )
        if res.status_code not in (200, 400):
            raise Exception('Status code {}, message {} for {}'.format(
                res.status_code, res.text, asset_id))

        if res.status_code == 400:
            return None
        data = res.json()
        if not data:
            return None
        return data[0]['name']

    def get_hdd_url(self, asset_id):
        return urljoin(self.base_url, '/app/assets/disks/{}/'.format(
            asset_id))

    def get_hdd(self, asset_id):
        res = self.__get(
            'assets/hdd/?id={0}'.format(
                asset_id)
        )
        if res.status_code == 200:
            data = res.json()
            if not data:
                return None
            return data
        elif res.status_code == 400:
            return None
        else:
            raise Exception('Status code {0}, message {1}'.format(
                res.status_code, res.text))

    def change_bay(self, hdd_id, bay):
        return self.__generic_patch(
            'assets/hdd/{0}/?update_bay'.format(hdd_id),
            {'bay': bay})

    def add_owner(self, hdd_id, name, email=' '):
        return self.__generic_patch(
            'assets/hdd/{0}/?update_owner'.format(hdd_id),
            {'name': name,
             'email': email})

    def add_location(self, hdd_id, location):
        return self.__generic_patch(
            'assets/hdd/{0}/?update_location'.format(hdd_id),
            {'location': location})

    def add_status(self, hdd_id, status, extra_info=' '):
        return self.__generic_patch(
            'assets/hdd/{0}/?update_status'.format(hdd_id),
            {'status': status,
             'extra_info': extra_info})

    def add_health_status(self, hdd_id, status, extra_info=' '):
        return self.__generic_patch(
            'assets/hdd/{0}/?update_health'.format(hdd_id),
            {'status': status,
             'extra_info': extra_info})

    def add_wbpclass(self, hdd_id, wbp_class_number, extra_info=' '):
        return self.__generic_patch(
            'assets/hdd/{0}/?update_wbpclass'.format(hdd_id),
            {'wbp_class_number': wbp_class_number,
             'extra_info': extra_info})

    def add_smart_data(self, hdd_id, rawdata):
        return self.__generic_post(
            'assets/hdd_smartdata/',
            {'data': rawdata,
             'hdd': hdd_id})

    def add_hdd(self, hdd_dock, bay='', asset_uid=None):
        data = {
            'device_model': hdd_dock.get_device_model(),
            'serial_number': hdd_dock.get_serial_number(),
            'user_capacity': hdd_dock.get_user_capacity(),
            'ata_version': hdd_dock.get_ata_version(),
            'firmware_version': hdd_dock.get_firmware_version(),
            'smart_status': hdd_dock.get_smart_status(),
            'bay': bay,
        }

        if hdd_dock.get_model_family():
            data['model_family'] = hdd_dock.get_model_family()

        return self.__generic_post('api/assets/hdd/', data)

    def get_hdd_id(self, device_model, serial_number):
        assert '\n' not in str(device_model) + str(serial_number)
        assert '&' not in str(device_model) + str(serial_number)

        if device_model:
            res = self.__get(
                'assets/hdd/?serial_number={0}&device_model={1}'.format(
                    serial_number, device_model))
        else:
            res = self.__get(
                'assets/hdd/?limit=2&offset=0&overview=True&search={}'.format(
                    serial_number))
            if res.status_code != 200:
                raise Exception('Status code {0}, message {1}'.format(
                    res.status_code, res.text))
            data = res.json()
            if data['count'] > 1:
                raise Exception('Multiple results: {}'.format(data))
            if data['count'] == 0:
                return None
            return data['results'][0]['id']

        if res.status_code == 200:
            data = res.json()
            if not data:
                return None
            return data[0]['id']
        else:
            raise Exception('Status code {0}, message {1}'.format(
                res.status_code, res.text))



import json
from urllib.parse import urljoin

import requests

from . import settings
from .jwt_token_provider import get_jwt_token


__author__ = 'osso'


class LabelPrinterRESTClient(object):
    def __init__(self, base_url, jwt_token):
        self.base_url = base_url
        self.auth_header = {
            'Authorization': 'JWT {0}'.format(jwt_token)}

    def __post_json(self, url, data):
        headers = {'content-type': 'application/json'}
        headers.update(self.auth_header)
        res = requests.post(
            urljoin(self.base_url, url), data=json.dumps(data),
            headers=headers)
        return res

    def get(self, url):
        return self.__get(url)

    def __get(self, url):
        print(urljoin(self.base_url, url))
        res = requests.get(
            urljoin(self.base_url, url), headers=self.auth_header)
        return res

    def __generic_post(self, url, data):
        res = self.__post_json(url, data)
        if res.status_code == 201:
            return res.json()
        else:
            raise Exception(
                "Status code {0}, message {1}".format(
                    res.status_code,
                    res.text))

    def print_generic_label(self, lines):
        return self.__generic_post(
            'api/text_label/',
            {'text': "\n".join(lines)})

    def print_hdd_label(self, hdd_uid, serial_number, owner=''):
        return self.__generic_post(
            'api/hdd_label/',
            {'hdd_uid': hdd_uid,
             'serial_number': serial_number,
             'owner': owner})

    def print_server_label(self, server_uid, name, owner=''):
        return self.__generic_post(
            'api/server_label/',
            {'server_uid': server_uid,
             'name': name,
             'owner': owner})


def get_labelprinter():
    jwt_token = get_jwt_token(
        settings.LABELPRINTER_LOGIN_URL,
        auth_token=settings.LABELPRINTER_AUTH_TOKEN)

    rst = LabelPrinterRESTClient(
        base_url=settings.LABELPRINTER_BASE_URL,
        jwt_token=jwt_token)

    return rst


if __name__ == '__main__':
    rst = get_labelprinter()
    res = rst.get('/api/text_label/')
    assert res.status_code == 405, (res.status_code, res)  # expected
    print(res)



from __future__ import print_function

from warnings import warn
import os.path
import re
import subprocess

from .devtype import DEVTYPE_NVME

__author__ = 'osso'


def _assert_int(value):
    "Precondition or postcondition to check that this is an int"
    assert isinstance(value, int), value
    return value


class NvmeDataError(Exception):
    pass


class NvmeDataParser(object):
    devtype = DEVTYPE_NVME

    def __init__(self, devname):
        self.devname = devname
        # Yes, nvme cli has --output-format=json too, but it lacks explanation
        # of which fields are what, and there is no guarantee that it won't
        # change either (and string fields are right-padded):
        # -H = --human-readable
        self.rawdata = (
            self._get_nvme_data(['id-ctrl', '-H']) + '\n' +
            self._get_nvme_data(['id-ns', '-H']) + '\n' +
            self._get_nvme_data(['smart-log']))
        self.information, self.attributes = (
            self._parse_nvme_information(self.rawdata))

    def _get_nvme_data(self, cmd):
        nvmecli = '/usr/sbin/nvme'
        try:
            p = subprocess.Popen(
                [nvmecli] + cmd + [self.devname], env={'LC_ALL': 'C'},
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = p.communicate()
        except FileNotFoundError as e:
            raise NvmeDataError('{} is not installed'.format(
                nvmecli)) from e

        # Some devices return junk as part of the vendor name.
        stdout = stdout.decode('utf-8', 'ignore')
        stderr = stderr.decode('utf-8', 'ignore')

        if p.returncode != 0:
            raise NvmeDataError('nvme cli error {} on {}: {}'.format(
                p.returncode, self.devname, (
                    stderr + '\n\n' + stdout).strip()))

        return stdout

    @property
    def device_model(self):
        """
        mn : INTEL SSDPE2KX010T8
        """
        return self.information['mn']

    @property
    def model_family(self):
        """
        ???
        """
        return None

    @property
    def serial_number(self):
        """
        sn : BTLM1170116EDJ94GK
        """
        return self.information['sn']

    @property
    def user_capacity(self):
        """
        nvmcap  : 1000204886016 <-- from 'id-ns'
        """
        # Caller expects string
        if self.information['nvmcap']:
            return str(self.information['nvmcap'])
        elif self.information['nsze']:
            return str(self.information['nsze'] * self.sector_size)
        raise NotImplementedError()

    @property
    def firmware_version(self):
        """
        fr : VDV10131
        """
        return self.information['fr']

    @property
    def ata_version(self):
        """
        ???
        """
        return 'unknown'

    @property
    def smart_status(self):
        """
        Must be enabled, since we didn't die on the 'smart-log' cmd..
        """
        assert isinstance(self.information['critical_warning'], int), (
            self.information['critical_warning'])
        if self.information['critical_warning']:
            return 'FAILED!'
        return 'PASSED'

    @property
    def sector_size(self):
        """
        LBA Format  0 : Metadata Size: 0   bytes - Data Size: 512 bytes -
          Relative Performance: 0x2 Good (in use)
        LBA Format  1 : Metadata Size: 0   bytes - Data Size: 4096 bytes -
          Relative Performance: 0 Best

        or, in json:

        "flbas" : 0  # "Formatted LBA Size" (current)
        "nlbaf" : 1  # 0-based count of sizeof(lbafs)
        "lbafs" : [
          {"ms" : 0, "ds" : 9, "rp" : 2},  # 9=512
          {"ms" : 0, "ds" : 12, "rp" : 0}  # 12=4096
        ]
        """
        if not hasattr(self, '_sector_size'):
            for key, value in self.information.items():
                if key.startswith('LBA Format ') and '(in use)' in value:
                    for line in value.split(' - '):
                        if line.startswith('Data Size'):
                            self._sector_size = int(
                                line.split(':', 1)[1].strip().split(' ', 1)[0])
                            return self._sector_size
                    return ValueError('Parse error of sector_size {!r}'.format(
                        value))
            return ValueError('Cannot find LBA Format in information')
        return self._sector_size

    @property
    def wear_health_percent(self):
        """
        Returns estimated health of device (from 100 down to 0 (or below))
        """
        # We want a value that represents the health, like the
        # wear_leveling_count did. Returns a signed integer from 100
        # and downwards. (May become negative.)
        used_health = int(self.information['percentage_used'].replace('%', ''))
        return 100 - used_health

    @property
    def power_on_hours(self):
        return _assert_int(self.information['power_on_hours'])

    @property
    def lbas_read(self):
        return _assert_int(self.information['host_read_commands'])

    @property
    def lbas_written(self):
        return _assert_int(self.information['host_write_commands'])

    @property
    def reallocated_sector_ct(self):
        return None  # ??? perhaps something with available_spare(_threshold)

    @property
    def reallocated_event_count(self):
        return None  # ??? see above

    @property
    def current_pending_sector(self):
        return None

    @property
    def offline_uncorrectable(self):
        return None

    def _parse_nvme_information(self, rawdata):
        it = iter(rawdata.split('\n'))

        # frmw      : 0x2
        #   [4:4] : 0   Firmware Activate Without Reset Not Supported
        #   [3:1] : 0x1	Number of Firmware Slots
        #   [0:0] : 0   Firmware Slot 1 Read/Write
        # ...
        #   [31:1] : 0x6a\tReserved
        bitre = re.compile(r'\s*\[(\d+):(\d+)\]\s*:\s*([0-9a-fx]+)\s+(.*)$')

        information = {}
        last_key = None

        for line in it:
            indent = line.startswith(' ')
            try:
                key, value = line.split(':', 1)
            except ValueError:
                # No ':' found.. ignore.
                pass
            else:
                key = key.strip()

                if key.startswith('['):
                    # [31:30] : 0x2 Blah <-- bits 30+31 have 31 set
                    self._populate_subinfo(
                        information, key, value, last_key, line, bitre)
                elif not indent:
                    last_key = key
                    self._populate_info(information, key, value)

        attributes = {}
        if not attributes:
            warn('No attributes found')

        return information, attributes

    def _populate_info(self, information, key, value):
        # "frmw      : 0x2" => {"frmw": 0x02}
        value = value.strip()
        if value.startswith('0x') or value == '0':
            value = int(value, 16)  # 0x12 -> 18
        elif all(group.isdigit() for group in value.split(',')) and (
                all(len(group) == 3 for group in value.split(',')[1:])):
            value = int(value.replace(',', ''))  # 1,234,567 -> 1234567
        if key not in information:
            information[key] = value
        else:
            assert False, ((key, value), information)

    def _populate_subinfo(self, information, key, value, last_key, line,
                          bitre):
        assert last_key  # must be something
        # > [31:30] : 0x2 Blah <-- bits 30+31 have 31 set
        # " [31:30] : 0x2 Blah" => {"frmw:30:31": 0x2}
        m = bitre.match(line)
        assert m, (
            'does not match bitre: {}'.format(bitre.pattern), line)
        bitend, bitstart, value, name = m.groups()
        if bitend == bitstart:
            value = bool(int(value, 16))    # 0, 0x1
            information['{}:{}'.format(last_key, bitend)] = value
        else:
            value = int(value, 16)          # 0, 0x1, 0x2, ...
            information['{}:{}:{}'.format(
                last_key, bitend, bitstart)] = value


if __name__ == "__main__":
    import sys

    if len(sys.argv) != 2:
        print('Usage {} device'.format(
            os.path.basename(sys.argv[0])), file=sys.stderr)
        sys.exit(1)

    # (in smart-log)
    # critical_warning                    : 0
    # temperature                         : 34 C
    # available_spare                     : 100%
    # available_spare_threshold           : 10%
    # percentage_used                     : 0%
    # data_units_read                     : 37
    # data_units_written                  : 0
    # host_read_commands                  : 1056
    # host_write_commands                 : 0
    # controller_busy_time                : 0
    # power_cycles                        : 4
    # power_on_hours                      : 912
    # unsafe_shutdowns                    : 2
    # media_errors                        : 0
    # num_err_log_entries                 : 0
    # Warning Temperature Time            : 0
    # Critical Composite Temperature Time : 0
    # Thermal Management T1 Trans Count   : 0
    # Thermal Management T2 Trans Count   : 0
    # Thermal Management T1 Total Time    : 0
    # Thermal Management T2 Total Time    : 0

    parser = NvmeDataParser(sys.argv[1])
    for prop in (
            'device_model',
            'model_family',
            'serial_number',
            'user_capacity',
            'firmware_version',
            'ata_version',
            'smart_status',
            'sector_size',
            'lbas_written',
            'lbas_read',
            'power_on_hours',
            'wear_health_percent',
            'reallocated_sector_ct',
            'reallocated_event_count',
            'current_pending_sector',
            'offline_uncorrectable'):
        val = getattr(parser, prop)
        str_val = str(val) if val is not None else '(N/A)'
        print('{:24}: {}'.format(prop, str_val))



# osso-docktool -- HDD administration and maintenance
# Copyright (C) 2015-2022 OSSO B.V.

from setuptools import setup, find_packages
from os.path import dirname, join

long_descriptions = []
with open(join(dirname(__file__), 'README.rst')) as file:
    long_descriptions.append(file.read())
version = '1.8.4'

setup(
    name='osso-docktool',
    version=version,
    data_files=[('share/doc/osso-docktool', [
        'README.rst', 'local_settings.py.template'])],
    entry_points={'console_scripts': [
        'osso-docktool = osso_docktool.docktool:main']},
    packages=find_packages(include=['osso_docktool', 'osso_docktool.*']),
    description='HDD administration and maintenance',
    long_description=('\n\n\n'.join(long_descriptions)),
    author='OSSO B.V.',
    author_email='dev+osso-docktool@osso.nl',
    url='https://git.osso.nl/osso-io/docktool',  # osso-int[ernal]?
    license='Undecided',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: System Administrators',
        # ('License :: OSI Approved :: GNU General Public License v3 '
        #  'or later (GPLv3+)'),
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: System :: Filesystems',
        'Topic :: Utilities',
    ],
    install_requires=[
        'requests',
    ],
)

# vim: set ts=8 sw=4 sts=4 et ai tw=79:



import sys

__author__ = 'osso'

DEBUG = False

LABELPRINTER_BASE_URL = 'https://labelprinter.example.com/'
LABELPRINTER_LOGIN_URL = 'https://labelprinter.example.com/api-auth/login/'
LABELPRINTER_AUTH_TOKEN = '<key-here>'

DASHBOARD_BASE_URL = 'https://dashboard.example.com/'
DASHBOARD_SCHEME_URL = 'https://dashboard.example.com/api/auth/scheme/'
DASHBOARD_AUTH_TOKEN = '<key-here>'

ATA_TO_BAY_NR = {
    'ata-6': 1,
    'ata-5': 2,
    'ata-4': 3,
    'ata-3': 4,
}

# List of devices that are ignored for udev/auto notifications.
IGNORE_DEVICES = []

# disk empty sample settings
# - divide disk in sample_count regions
# - sample sample_size bytes randomly per region
POST_WIPE_SAMPLE_COUNT = 1000
POST_WIPE_SAMPLE_SIZE = 100 * 1024

# Find local_settings.py in /etc/osso-docktool or in the parent directory.
sys.path.insert(0, '/etc/osso-docktool')
try:
    from local_settings import *  # noqa
except ImportError:
    pass
